# Copyright 2023-2024 Broadcom
# SPDX-License-Identifier: Apache-2.0
import json
import logging
import time
from typing import Optional

from requests import post
from requests.auth import HTTPBasicAuth
from requests_oauthlib import OAuth2Session
from vdk.plugin.control_api_auth.auth_config import AuthConfig
from vdk.plugin.control_api_auth.auth_config import CredentialsCache
from vdk.plugin.control_api_auth.auth_config import LocalFolderCredentialsCache
from vdk.plugin.control_api_auth.auth_exception import VDKAuthException
from vdk.plugin.control_api_auth.auth_request_values import AuthRequestValues
from vdk.plugin.control_api_auth.login_types import LoginTypes

log = logging.getLogger(__name__)


class AuthenticationCache:
    """
    Class used to keep the fields in the vdk-cred.json file from which
    authorization and authentication tokens can be taken. Used only when
    caching is enabled.
    """

    def __init__(
        self,
        authorization_url=None,
        refresh_token=None,
        access_token=None,
        access_token_expiration_time=0,
        client_id=None,
        client_secret=None,
        api_token=None,
        api_token_authorization_url=None,
        auth_type=None,
        **ignored_kwargs,
    ):
        """
        :param authorization_url:
            Token or Authorization URI used to exchange grant for access token
        :param refresh_token:
            Used in refresh token grant type flow;
            See https://datatracker.ietf.org/doc/html/rfc6749#section-1.5
        :param access_token:
            Cached access token used during requests;
            See https://datatracker.ietf.org/doc/html/rfc6749#section-1.4
        :param access_token_expiration_time:
            When the access token expires
            (epoch seconds -- same as time.time() + "expires_in" value returned
            by the authorization server that generates the access token.)
        :param client_id:
            The client identifier;
            See https://datatracker.ietf.org/doc/html/rfc6749#section-2.3.1
        :param api_token:
            Token used to grant access to the API,
            including exchange for access token.
        :param api_token_authorization_url:
            The URL which exchanges api_token for access token.
        """
        #
        self.authorization_url = authorization_url
        self.refresh_token = refresh_token
        self.access_token = access_token
        self.access_token_expiration_time = access_token_expiration_time
        self.client_id = client_id
        self.client_secret = client_secret
        self.api_token = api_token
        self.api_token_authorization_url = api_token_authorization_url
        self.auth_type = auth_type

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class AuthenticationCacheSerDe:
    """Class used for cache serialization and deserialization."""

    @staticmethod
    def serialize(cache) -> str:
        return json.dumps(cache.__dict__)

    @staticmethod
    def deserialize(content: str) -> AuthenticationCache:
        if content:
            return AuthenticationCache(**json.loads(content))
        else:
            return AuthenticationCache()


class BaseAuth:
    REFRESH_TOKEN_GRANT_TYPE = "refresh_token"  # nosec

    def __init__(
        self,
        conf: CredentialsCache = LocalFolderCredentialsCache(),
        vdk_config=AuthConfig(),
    ):
        self._conf = conf
        self.__vdk_config = vdk_config
        self._cache = self.__load_cache()

    def __load_cache(self):
        content = self._conf.read_credentials()
        return AuthenticationCacheSerDe.deserialize(content)

    def __update_cache(self):
        self._conf.save_credentials(AuthenticationCacheSerDe.serialize(self._cache))

    def __exchange_api_for_access_token(self):
        try:
            log.debug(
                f"Refresh API token against {self._cache.api_token_authorization_url} "
            )
            client = OAuth2Session()
            token_response = client.refresh_token(
                self._cache.api_token_authorization_url,
                refresh_token=self._cache.api_token,
            )
            log.debug(
                f"Token response received: "
                f"token_type: {token_response.get('token_type', None)}; "
                f"expires_in: {token_response.get('expires_in', None)}; "
                f"scope: {token_response.get('scope', None)}"
            )
        except Exception as e:
            raise VDKAuthException(
                what="Failed to login",
                why=f"Authorization server at {self._cache.api_token_authorization_url} returned error: {str(e)}",
                consequence="Your credentials are not refreshed and VDK control operations that require authentication "
                "will not work.",
                countermeasure="Check error message and follow instructions in it. Check your network connectivity."
                " Check if you can access the Oauth2 server. Retry to login a few times again."
                " Contact SRE Team operating Oauth2 server for help if nothing else works.",
            ) from e
        # Response per spec in https://tools.ietf.org/html/rfc6750#section-4
        # see also https://www.oauth.com/oauth2-servers/access-tokens/access-token-response
        self.update_access_token(
            token_response[AuthRequestValues.ACCESS_TOKEN_KEY.value]
        )
        self.update_access_token_expiration_time(
            time.time()
            + int(token_response.get(AuthRequestValues.EXPIRATION_TIME_KEY.value, "0"))
        )

    def __exchange_refresh_for_access_token(self):
        basic_auth = HTTPBasicAuth(self._cache.client_id, self._cache.client_secret)
        headers = {
            AuthRequestValues.CONTENT_TYPE_HEADER.value: AuthRequestValues.CONTENT_TYPE_URLENCODED.value,
        }
        data = (
            f"grant_type={self.REFRESH_TOKEN_GRANT_TYPE}&"
            + f"refresh_token={self._cache.refresh_token}"
        )
        log.debug(
            f"Refresh access token against {self._cache.authorization_url} grant_type={self.REFRESH_TOKEN_GRANT_TYPE}"
        )
        response = post(
            self._cache.authorization_url, data=data, headers=headers, auth=basic_auth
        )
        if response.status_code >= 400:
            raise VDKAuthException(
                what="Failed to refresh access token",
                why="Authorization server returned status code: {response.status_code}, error response: {response.text}",
                consequence="Subsequent API calls may fail.",
                countermeasure="Try to log in again the usual way. "
                "If using the CLI, you can login using the 'vdk login' command. "
                "If using a notebook, use the Login Button.",
            )

        json_data = json.loads(response.text)
        log.debug(
            f"Refresh access token response received: status: {response.status_code}"
        )
        self.update_access_token(json_data[AuthRequestValues.ACCESS_TOKEN_KEY.value])
        self.update_access_token_expiration_time(
            time.time() + int(json_data[AuthRequestValues.EXPIRATION_TIME_KEY.value])
        )

    def read_access_token(self) -> Optional[str]:
        """
        Read access token from cache or fetch it from Authorization server.
        If not available in cache it will get it using provided configuration during VDK CLI login to fetch it.
        If it detects that token is about to expire it will try to refresh it.
        :return: the access token or None if it cannot detect any credentials.
        """
        if (
            not self._cache.access_token
            or self._cache.access_token_expiration_time < time.time() + 60
        ):
            log.debug("Acquire access token (it's either expired or missing) ...")
            self.acquire_and_cache_access_token()
        return self._cache.access_token

    def read_cached_access_token_only(self) -> Optional[str]:
        """
        Read access token from credentials cache and will return it if available.
        Will not check if it is valid (for example not expired) but return it directly.
        Will not try to fetch it form Authorization provider if it is not available. Instead, will simply return None
        """
        return self._cache.access_token

    def acquire_and_cache_access_token(self):
        """
        Acquires and caches access token
        """
        log.debug(f"Using auth type {self._cache.auth_type} to acquire access token")
        if (
            self._cache.auth_type == LoginTypes.API_TOKEN.value
            and self._configured_api_token()
        ):
            self.__exchange_api_for_access_token()
        elif (
            self._cache.auth_type == LoginTypes.CREDENTIALS.value
            and self._configured_refresh_token()
        ):
            self.__exchange_refresh_for_access_token()
        elif (
            self.__vdk_config.api_token
            and self.__vdk_config.api_token_authorization_url
        ):
            self.update_api_token(self.__vdk_config.api_token)
            self.update_api_token_authorization_url(
                self.__vdk_config.api_token_authorization_url
            )
            self.__exchange_api_for_access_token()
        else:
            log.debug(
                "No authentication mechanism found. Will not cache access token."
                "If Control Service authentication is enabled, API calls will fail."
            )

    def update_api_token_authorization_url(self, api_token_authorization_url):
        """
        Updates and caches authorization URL
        """
        self._cache.api_token_authorization_url = api_token_authorization_url
        self.__update_cache()

    def update_api_token(self, api_token):
        """
        Updates and caches refresh token
        """
        self._cache.api_token = api_token
        self.__update_cache()

    def update_refresh_token(self, refresh_token):
        """
        Updates and caches refresh token
        """
        self._cache.refresh_token = refresh_token
        self.__update_cache()

    def update_client_id(self, client_id):
        """
        Updates and caches refresh token
        """
        self._cache.client_id = client_id
        self.__update_cache()

    def update_client_secret(self, client_secret):
        """
        Updates and caches refresh token
        """
        self._cache.client_secret = client_secret
        self.__update_cache()

    def update_access_token(self, access_token):
        """
        Updates and caches refresh token
        """
        self._cache.access_token = access_token
        self.__update_cache()

    def update_access_token_expiration_time(self, access_token_expiration_time):
        """
        Updates and caches refresh token
        """
        self._cache.access_token_expiration_time = access_token_expiration_time
        self.__update_cache()

    def update_oauth2_authorization_url(self, authorization_url):
        self._cache.authorization_url = authorization_url
        self.__update_cache()

    def update_auth_type(self, auth_type):
        self._cache.auth_type = auth_type
        self.__update_cache()

    def _configured_api_token(self):
        return self._cache.api_token and self._cache.api_token_authorization_url

    def _configured_refresh_token(self):
        return (
            self._cache.refresh_token
            and self._cache.authorization_url
            and self._cache.client_id
            and self._cache.client_secret
        )



# Copyright 2023-2024 Broadcom
# SPDX-License-Identifier: Apache-2.0
import logging
import warnings
from typing import Optional

from vdk.plugin.control_api_auth.auth_config import CredentialsCache
from vdk.plugin.control_api_auth.auth_config import LocalFolderCredentialsCache
from vdk.plugin.control_api_auth.auth_config import SingletonInMemoryCredentialsCache
from vdk.plugin.control_api_auth.auth_exception import VDKInvalidAuthParamError
from vdk.plugin.control_api_auth.autorization_code_auth import RedirectAuthentication
from vdk.plugin.control_api_auth.base_auth import BaseAuth
from vdk.plugin.control_api_auth.login_types import LoginTypes

log = logging.getLogger(__name__)


class Authentication:
    """Main class used for authentication."""

    def __init__(
        self,
        username: str = None,
        password: str = None,
        client_id: str = None,
        client_secret: str = None,
        token: str = None,
        authorization_url: str = None,
        auth_discovery_url: str = None,
        auth_type: str = None,
        cache_locally: bool = None,
        possible_jwt_user_keys=None,
        credentials_cache: CredentialsCache = LocalFolderCredentialsCache(),
    ):
        """
        :param username: A user's username in case basic authentication is used.
        :param password: A user's password in case basic authentication is used.
        :param client_id:
            The client identifier;
            See https://datatracker.ietf.org/doc/html/rfc6749#section-2.3.1
        :param client_secret:
            The client identifier;
            See https://datatracker.ietf.org/doc/html/rfc6749#section-2.3.1
        :param token:
            OAuth api token or a refresh token as specified in
            https://datatracker.ietf.org/doc/html/rfc6749#section-1.5 and used
            to obtain access token from the authorization server.
        :param authorization_url:
            The URL which exchanges token for access token.
        :param auth_discovery_url:
            Token or Authorization URI used to exchange grant for access token.
        :param auth_type:
            What type of authentication should be used (e.g., refresh_token,
            basic_auth, etc.).
        :param cache_locally:
            A flag, indicating if credentials should be cached locally (in a
            file).
            Deprecated: use credentials_cache instead to pass how credentials are stored explicitly.
        :param possible_jwt_user_keys:
            Used by get_authenticated_username to try to discover correct username if OAuth2 and JWT token is used. It is a list of keys where the first existing key in a JTW token is returned.
            Defaults to some common user keys.
        :param credentials_cache:
            The configuration store used to actually store the credentials
        """
        self._username = username
        self._password = password
        self._client_id = client_id
        self._client_secret = client_secret
        self._token = token
        self._auth_url = authorization_url
        self._auth_discovery_url = auth_discovery_url
        self._auth_type = auth_type

        if cache_locally is not None:
            warnings.warn(
                "Authentication constructor cache_locally argument is deprecated. "
                "Use credentials_cache to pass explicitly the credential cache store."
            )
            credentials_cache = (
                LocalFolderCredentialsCache()
                if cache_locally
                else SingletonInMemoryCredentialsCache()
            )

        self._auth = BaseAuth(credentials_cache)

        if possible_jwt_user_keys:
            self._possible_jwt_user_keys = possible_jwt_user_keys
        else:  # sensible defaults
            self._possible_jwt_user_keys = [
                "username",
                "acct",
                "preferred_username",
                "email",
            ]

    def authenticate(self) -> None:
        if not self._auth_type:
            raise VDKInvalidAuthParamError(
                what="Unable to log in.",
                why="auth_type was not specified.",
                consequence="Subsequent requests to Control Service will not "
                " be authenticated.",
                countermeasure="Specify what type of authentication is to be " "used.",
            )
        if not self._auth_url:
            raise VDKInvalidAuthParamError(
                what="Unable to log in.",
                why="auth_url was not specified.",
                consequence="Authentication is not possible. All subsequent "
                "requests to Control Service will not be authenticated.",
                countermeasure="Provide a valid authorization url.",
            )

        if self._auth_type == LoginTypes.API_TOKEN.value:
            self.__authenticate_with_api_token()
        elif self._auth_type == LoginTypes.CREDENTIALS.value:
            self.__authenticate_with_authorization_code()
        else:
            raise VDKInvalidAuthParamError(
                what="Unexpected authentication type.",
                why=f"Unknown auth_type {self._auth_type} was used.",
                consequence="Authentication is not possible.",
                countermeasure="Provide a valid auth_type.",
            )

    def get_authenticated_username(self) -> Optional[str]:
        """
        Extract user name about currently authenticated user.
        It tries to get username or if not available user email or if that fails some user id.
        The reponse format should not be relied upon, it's meant to be used for logging and telemetry.
        """
        if self._username:
            return self._username

        if not self._token:
            try:
                self._token = self.read_access_token()
            except Exception:
                log.debug(f"Could not to read access token.", exc_info=True)
        if not self._token:
            self._token = self._auth.read_cached_access_token_only()
        if not self._token:
            return None

        try:
            import jwt

            jwt_payload = jwt.decode(self._token, options={"verify_signature": False})
            if not jwt_payload:
                return None

            for key in self._possible_jwt_user_keys:
                user_id = jwt_payload.get(key)
                if user_id:
                    return user_id

        except Exception:
            log.debug(
                f"Could not to extract user information from the token.", exc_info=True
            )
            return None

    def read_access_token(self):
        """Read access token from cache."""
        return self._auth.read_access_token()

    def __authenticate_with_api_token(self):
        """Authenticate by providing only a OAuth2 API token."""
        self._auth.update_api_token_authorization_url(
            api_token_authorization_url=self._auth_url
        )
        self._auth.update_api_token(api_token=self._token)
        self._auth.update_auth_type(auth_type=LoginTypes.API_TOKEN.value)
        self._auth.acquire_and_cache_access_token()

    # NOTE: Implementation to be added with subsequent PR
    def __authenticate_with_authorization_code(self):
        """
        Authenticate with authorization code as described in
        https://datatracker.ietf.org/doc/html/rfc6749#section-1.3.1
        This type of authentication relies on pkce
        https://datatracker.ietf.org/doc/html/rfc7636
        :return:
        """
        if not self._client_id or not self._auth_discovery_url:
            raise VDKInvalidAuthParamError(
                what="Unable to log in.",
                why=f"Login type {self._auth_type} requires client_id and "
                "auth_discovery_url to be specified.",
                consequence="The authentication operation cannot be complete.",
                countermeasure="Specify client_id and auth_discovery_url and "
                "repeat the login operation.",
            )
        else:
            auth_code_flow = RedirectAuthentication(
                client_id=self._client_id,
                client_secret=self._client_secret,
                oauth2_discovery_url=self._auth_discovery_url,
                oauth2_exchange_url=self._auth_url,
                auth=self._auth,
            )
            auth_code_flow.authentication_process()



# Copyright 2023-2024 Broadcom
# SPDX-License-Identifier: Apache-2.0
from base64 import urlsafe_b64encode

import pytest
from pytest_httpserver.pytest_plugin import PluginHTTPServer
from test_core_auth import allow_oauthlib_insecure_transport
from test_core_auth import get_json_response_mock
from vdk.plugin.control_api_auth.auth_config import InMemoryCredentialsCache
from vdk.plugin.control_api_auth.auth_exception import VDKInvalidAuthParamError
from vdk.plugin.control_api_auth.auth_exception import VDKLoginFailedError
from vdk.plugin.control_api_auth.authentication import Authentication
from vdk.plugin.control_api_auth.autorization_code_auth import generate_pkce_codes
from vdk.plugin.control_api_auth.autorization_code_auth import LoginHandler
from vdk.plugin.control_api_auth.autorization_code_auth import RedirectAuthentication
from vdk.plugin.control_api_auth.base_auth import BaseAuth


def test_verify_redirect_url():
    allow_oauthlib_insecure_transport()
    in_mem_conf = InMemoryCredentialsCache()
    auth = BaseAuth(in_mem_conf)

    auth = RedirectAuthentication(
        client_id="client-id",
        client_secret="client-secret",
        oauth2_discovery_url="http://discovery-url",
        oauth2_exchange_url="http://exchange-url",
        auth=auth,
        redirect_uri="http://127.0.0.1",
        redirect_uri_default_port=9999,
    )
    authorization_url = auth._create_authorization_redirect_url()
    assert (
        authorization_url[0]
        == "http://discovery-url?response_type=code&client_id=client-id&redirect_uri=http%3A%2F%2F127.0.0.1%3A9999&state=requested&prompt=login"
    )
    assert authorization_url[1] == "requested"


def test_login_handler_exceptions(httpserver: PluginHTTPServer):
    allow_oauthlib_insecure_transport()
    httpserver.expect_request("/foo").respond_with_json(get_json_response_mock())
    in_mem_conf = InMemoryCredentialsCache()
    auth = BaseAuth(in_mem_conf)

    handler = LoginHandler(
        client_id=None,
        client_secret=None,
        exchange_endpoint=httpserver.url_for("/foo"),
        redirect_uri=None,
        code_verifier=None,
        auth=auth,
    )

    with pytest.raises(VDKLoginFailedError) as exc_info:
        handler.login_with_authorization_code(
            path="http://test-url?code=test-auth-code&client_id=client-id&redirect_uri=http%3A%2F%2F127.0.0.1%3A9999&prompt=login"
        )
    raised_exception = exc_info.value
    assert "Failed to login." in raised_exception.message
    assert "Possibly the request was intercepted." in raised_exception.message

    with pytest.raises(VDKLoginFailedError) as exc_info2:
        handler.login_with_authorization_code(
            path="http://test-url?client_id=client-id&redirect_uri=http%3A%2F%2F127.0.0.1%3A9999&state=requested&prompt=login"
        )
    raised_exception2 = exc_info2.value
    assert "Authentication code is empty" in raised_exception2.message
    assert "The user failed to authenticate properly." in raised_exception2.message


def test_authorization_code_no_secret(httpserver: PluginHTTPServer):
    allow_oauthlib_insecure_transport()
    httpserver.expect_request("/foo").respond_with_json(get_json_response_mock())

    auth = Authentication(
        token="apitoken",
        authorization_url=httpserver.url_for("/foo"),
        auth_type="credentials",
        credentials_cache=InMemoryCredentialsCache(),
    )
    with pytest.raises(VDKInvalidAuthParamError) as exc_info:
        auth.authenticate()

    raised_exception = exc_info.value
    assert "Unable to log in." in raised_exception.message
    assert "Specify client_id and auth_discovery_url" in raised_exception.message


@pytest.fixture
def mock_post_req(requests_mock):
    url = "http://example.com/test"
    requests_mock.post(url, json={"result": "Success"})


def test_authorization_code_authorization_header(mock_post_req, requests_mock):
    allow_oauthlib_insecure_transport()
    in_mem_conf = InMemoryCredentialsCache()
    auth = BaseAuth(in_mem_conf)
    (
        code_verifier,
        _,
        _,
    ) = generate_pkce_codes()
    auth_code = "someDummyAuthCode"
    auth_base64_string = urlsafe_b64encode(bytes("test-client-id:", "utf-8"))
    expected_authorization_header_value = f"Basic {auth_base64_string.decode('utf-8')}"

    handler = LoginHandler(
        client_id="test-client-id",
        client_secret=None,
        exchange_endpoint="http://example.com/test",
        redirect_uri="127.0.0.1:31113",
        code_verifier=code_verifier,
        auth=auth,
    )

    handler._exchange_code_for_tokens(auth_code)

    assert "Authorization" in requests_mock.last_request.headers
    assert (
        requests_mock.last_request.headers["Authorization"]
        == expected_authorization_header_value
    )



# Copyright 2023-2024 Broadcom
# SPDX-License-Identifier: Apache-2.0
from unittest.mock import patch

import pytest
from vdk.plugin.control_api_auth.auth_config import InMemoryCredentialsCache
from vdk.plugin.control_api_auth.authentication import Authentication


def test_get_authenticated_username_with_username():
    auth = Authentication(
        username="testuser", credentials_cache=InMemoryCredentialsCache()
    )
    assert auth.get_authenticated_username() == "testuser"


@patch("jwt.decode")
def test_get_authenticated_username_with_token_username(mock_jwt_decode):
    mock_jwt_decode.return_value = {"username": "testuser"}
    auth = Authentication(
        token="testtoken", credentials_cache=InMemoryCredentialsCache()
    )
    assert auth.get_authenticated_username() == "testuser"


@pytest.mark.parametrize("user_id_field", ["acct", "email", "preferred_username"])
@patch("jwt.decode")
def test_get_authenticated_username_with_token_other_fields(
    mock_jwt_decode, user_id_field
):
    mock_jwt_decode.return_value = {user_id_field: "testuser"}
    auth = Authentication(
        token="testtoken", credentials_cache=InMemoryCredentialsCache()
    )
    assert auth.get_authenticated_username() == "testuser"


@patch("jwt.decode")
def test_get_authenticated_username_with_token_no_user_id(mock_jwt_decode):
    mock_jwt_decode.return_value = {}
    auth = Authentication(
        token="testtoken", credentials_cache=InMemoryCredentialsCache()
    )
    assert auth.get_authenticated_username() is None


@patch("jwt.decode")
def test_get_authenticated_username_priority_order_kept(mock_jwt_decode):
    mock_jwt_decode.return_value = {"sub": "user_from_sub", "acct": "user_from_acct"}
    auth = Authentication(
        token="testtoken",
        credentials_cache=InMemoryCredentialsCache(),
        possible_jwt_user_keys=["user", "email", "sub", "acct"],
    )
    assert auth.get_authenticated_username() == "user_from_sub"


def test_get_authenticated_username_without_username_or_token():
    auth = Authentication(credentials_cache=InMemoryCredentialsCache())
    assert auth.get_authenticated_username() is None
